import typing as t

import networkx as nx
import typer

from paraffin.submit import submit_node_graph
from paraffin.utils import (
    dag_to_levels,
    get_custom_queue,
    get_stage_graph,
    levels_to_mermaid,
)

app = typer.Typer()


@app.command()
def main(
    names: t.Optional[list[str]] = typer.Argument(
        None, help="Stage names to run. If not specified, run all stages."
    ),
    concurrency: int = typer.Option(
        0,
        "--concurrency",
        "-c",
        envvar="PARAFFIN_CONCURRENCY",
        help="Number of stages to run in parallel. If not provided,"
        + " a celery worker has to be started manually.",
    ),
    glob: bool = typer.Option(
        False, "--glob", "-g", help="Use glob pattern to match stage names."
    ),
    shutdown_after_finished: bool = typer.Option(
        False,
        "--shutdown-after-finished",
        "-s",
        envvar="PARAFFIN_SHUTDOWN_AFTER_FINISHED",
        help="Shutdown the worker after all tasks are finished (experimental).",
    ),
    show_mermaid: bool = typer.Option(
        True, help="Visualize the parallel execution graph using Mermaid."
    ),
    dry: bool = typer.Option(False, help="Dry run. Do not submit tasks."),
):
    """Run DVC stages in parallel using Celery."""
    from paraffin.worker import app as celery_app

    graph = get_stage_graph(names=names, glob=glob)
    custom_queues = get_custom_queue()

    disconnected_subgraphs = list(nx.connected_components(graph.to_undirected()))
    disconnected_levels = [
        dag_to_levels(graph.subgraph(sg)) for sg in disconnected_subgraphs
    ]
    # iterate disconnected subgraphs for better performance
    if not dry:
        for levels in disconnected_levels:
            submit_node_graph(
                levels,
                shutdown_after_finished=shutdown_after_finished,
                custom_queues=custom_queues,
            )
    if show_mermaid:
        typer.echo(levels_to_mermaid(disconnected_levels))

    typer.echo(f"Submitted all (n = {len(graph)})  tasks.")
    if not dry:
        if concurrency > 0:
            celery_app.worker_main(
                argv=[
                    "worker",
                    "--loglevel=info",
                    f"--concurrency={concurrency}",
                    "--without-gossip",
                ]
            )
        else:
            typer.echo(
                "Start your celery worker using `celery -A paraffin.worker worker`"
                " and specify concurrency with `--concurrency`."
            )



import fnmatch
import typing as t

from celery import chain, group

from paraffin.abc import HirachicalStages
from paraffin.worker import repro, shutdown_worker


def submit_node_graph(
    levels: HirachicalStages,
    shutdown_after_finished: bool = False,
    custom_queues: t.Optional[t.Dict[str, str]] = None,
):
    per_level_groups = []
    for nodes in levels.values():
        group_tasks = []
        for node in nodes:
            if matched_pattern := next(
                (
                    pattern
                    for pattern in custom_queues
                    if fnmatch.fnmatch(node.name, pattern)
                ),
                None,
            ):
                group_tasks.append(
                    repro.s(name=node.name).set(queue=custom_queues[matched_pattern])
                )
            else:
                group_tasks.append(repro.s(name=node.name))
        per_level_groups.append(group(group_tasks))

    workflow = chain(per_level_groups)

    if shutdown_after_finished:
        chain(workflow, shutdown_worker.s()).apply_async()
    else:
        workflow.apply_async()



import logging
import pathlib
import subprocess
import time

from celery import Celery

log = logging.getLogger(__name__)


def make_celery() -> Celery:
    paraffin_folder = pathlib.Path(".paraffin")
    data_folder = paraffin_folder / "data"
    control_folder = paraffin_folder / "data"
    results_db = paraffin_folder / "results.db"
    gitignore = paraffin_folder / ".gitignore"

    data_folder.mkdir(parents=True, exist_ok=True)
    control_folder.mkdir(parents=True, exist_ok=True)
    results_db.parent.mkdir(parents=True, exist_ok=True)

    if not gitignore.exists():
        gitignore.write_text("data\nresults.db\n")

    app = Celery(
        __name__,
        broker_url="filesystem://",
        result_backend=f"db+sqlite:///{results_db.as_posix()}",
        broker_transport_options={
            "data_folder_in": data_folder.as_posix(),
            "data_folder_out": data_folder.as_posix(),
            "data_folder_processed": data_folder.as_posix(),
            "control_folder": control_folder.as_posix(),
        },
    )

    # app = Celery(
    #     __name__,
    #     broker="redis://localhost:6379/0",
    #     backend="redis://localhost:6379/0",
    # )

    return app


app = make_celery()


@app.task(bind=True, default_retry_delay=5)  # retry in 5 seconds
def repro(self, *args, name: str):
    popen = subprocess.Popen(
        ["dvc", "repro", "--single-item", name],
        stdout=subprocess.PIPE,
        universal_newlines=True,
        stderr=subprocess.PIPE,
    )
    for stdout_line in iter(popen.stdout.readline, ""):
        # logging.info(stdout_line)
        print(stdout_line, end="")
    popen.stdout.close()

    for stderr_line in iter(popen.stderr.readline, ""):
        # logging.error(stderr_line)
        print(stderr_line, end="")
        if "ERROR: Unable to acquire lock" in stderr_line:
            log.error(f"Retrying {name} due to lock error")
            raise self.retry(max_retries=5)
        if (
            f"ERROR: failed to reproduce '{name}': Unable to acquire lock"
            in stderr_line
        ):
            # unable to commit lock, keep retrying
            for _ in range(5):
                try:
                    log.error(f"Committing {name} again due to lock error")
                    subprocess.check_call(
                        ["dvc", "commit", name, "--force"],
                        stderr=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                    )
                    break
                except subprocess.CalledProcessError:
                    time.sleep(1)
            else:
                raise RuntimeError(f"Unable to commit lock for {name}")
    popen.stderr.close()
    return True


# Shutdown task
@app.task(bind=True)
def shutdown_worker(self, *args, **kwargs):
    app.control.revoke(self.request.id)  # prevent this task from being executed again
    app.control.shutdown()



import fnmatch
import pathlib

import dvc.api
import networkx as nx
import yaml

from paraffin.abc import HirachicalStages


def get_subgraph_with_predecessors(graph, nodes, reverse=False):
    # Initialize a set to store nodes that will be in the subgraph
    nodes_to_include = set(nodes)

    # For each node in X, find all its predecessors
    for node in nodes:
        predecessors = nx.ancestors(graph, node)
        nodes_to_include.update(predecessors)

    # Create the subgraph with the selected nodes
    if reverse:
        return graph.subgraph(nodes_to_include).reverse(copy=True)
    return graph.subgraph(nodes_to_include).copy()


def get_stage_graph(names, glob=False):
    fs = dvc.api.DVCFileSystem(url=None, rev=None)
    graph = fs.repo.index.graph.reverse(copy=True)

    nodes = [x for x in graph.nodes if hasattr(x, "name")]
    if names is not None:
        if glob:
            nodes = [
                x for x in nodes if any(fnmatch.fnmatch(x.name, name) for name in names)
            ]
        else:
            nodes = [x for x in nodes if x.name in names]

    subgraph = get_subgraph_with_predecessors(graph, nodes)

    # remove all nodes that do not have a name
    subgraph = nx.subgraph_view(subgraph, filter_node=lambda x: hasattr(x, "name"))

    return subgraph


def get_custom_queue():
    try:
        with pathlib.Path("paraffin.yaml").open() as f:
            config = yaml.safe_load(f)

            return config.get("queue", {})

    except FileNotFoundError:
        return {}


def dag_to_levels(graph) -> HirachicalStages:
    nodes = []
    levels = {}
    for start_node in graph.nodes():
        if len(list(graph.predecessors(start_node))) == 0:
            if start_node not in nodes:
                for node in nx.bfs_tree(graph, start_node):
                    if node not in nodes:
                        nodes.append(node)
                        level = nx.shortest_path_length(graph, start_node, node)
                        try:
                            levels[level].append(node)
                        except KeyError:
                            levels[level] = [node]
                    else:
                        # this part has already been added
                        break
    return levels


def levels_to_mermaid(all_levels: list[HirachicalStages]) -> str:
    # Initialize Mermaid syntax
    mermaid_syntax = "flowchart TD\n"

    for idx, levels in enumerate(all_levels):
        # Add each level as a subgraph
        for level, nodes in levels.items():
            mermaid_syntax += f"\tsubgraph Level{idx}:{level + 1}\n"
            for node in nodes:
                mermaid_syntax += f"\t\t{node.name}\n"
            mermaid_syntax += "\tend\n"

        # Add connections between levels
        for i in range(len(levels) - 1):
            mermaid_syntax += f"\tLevel{idx}:{i + 1} --> Level{idx}:{i + 2}\n"

    return mermaid_syntax
